// C Source File
// Created 20/10/00; 20:22:30

// Available Patches
// #define USE_GRAY_PATCH
// #define USE_LONGMUL_PATCH
// #define USE_LONGDIV_PATCH
#define USE_INITS_PATCH

#define SAVE_SCREEN  // Save/Restore LCD Contents
#define OPTIMIZE_ROM_CALLS



#include <nostub.h>  // Make NoStub Program
#include <graph.h>   // Support Graphics
#include <kbd.h>
#include <alloc.h>
#include <system.h>
#include <mem.h>
#include <dialogs.h>
#include <compat.h>
#include <stdio.h>


int _ti89, _ti92plus;           // Produce .89Z & .9XZ File

//const char current_version[]= "beta 3";

char* BL_sprites=NULL;
char* CHAR_sprites=NULL;

char* Curr_level= NULL;
volatile int keydown= 1;
int movekeydown= 1;

int Xpos= 0;
int Ypos= 0;

SCR_RECT *const full_screen= &(SCR_RECT){{0, 0, 239, 127}};


enum type_move {MV_UP, MV_RIGHT, MV_DOWN, MV_LEFT};
enum blocks { BL_VOID, BL_FIXE, BL_LEGER, BL_HEAVY, BL_HOLE, BL_REACT, BL_UP, BL_RIGHT, BL_DOWN, BL_LEFT, BL_HORIZ, BL_VERT, BL_EXPLODE, BL_JUMP, BL_TELEPORT, BL_WIN };

#include "bl_int1.c"
#include "getkey.c"
#include "bl_sprites.c"
#include "CHAR_sprites.c"
#include "extern_lvl.c"
#include "bl_intro.c"


char* level= NULL;
int Xsize= 0;
int Ysize= 0;
int to_win= 0;


void _main()
{
	void loadlevel();
	int bl_quit();
	void scroll();
	void getmove();
	
	asm("
		move.l #bl_sprites, BL_sprites
		move.l #char_sprites, CHAR_sprites
	");
	
	
	{
		int tmp=getlvl ();
		if (tmp!=1)
		{
			if (!tmp) DlgMessage ("Error", "No level file found!", BT_OK, BT_NONE);
			return;
		}
	}
	
	//desactive les auto-ints 1  4
	OSSetSR (0x300);
	
	Xscroll= 0;
	Yscroll= 0;
	
		
	if (!int1_on()) { return;}

	if (intro_screen()) { goto exit_1; }
	
	level= NULL;
	loadlevel();
		
	while (!bl_quit())	//main loop
	{
			if (!is_92) scroll();
			getmove();
	}
	
	free (level);
exit_1:
	int1_off();
	//ractive les auto-ints
	OSSetSR (0x0000);
	GKeyFlush();
	
	return;
}


void loadlevel()
{
	void copylevel (unsigned char* source);
	void drawlevel ();
	void drawchar (int type);
	
	copylevel (Curr_level);
	drawlevel ();
	drawchar (0);
}

void copylevel (unsigned char* source)
{
	int I, J;
	Xsize= *source;
	Ysize= *(source+1);
	to_win= *(source+2);
	Xpos= 0;
	Ypos= 0;
	Xscroll= 0;
	Yscroll= 0;
	if (level) { free(level); }
	level= malloc ((Xsize+1)/2*Ysize);
//	memset (level, 0, (Xsize+1)/2*Ysize);
	for (I=0; I<Ysize; I++)
	{
		for (J=0; J<((Xsize+1)/2); J++) { *(level+((Xsize+1)/2)*I+J)=*(source+((Xsize+1)/2)*I+J+HEADER_SIZE); }
	}
	
	return;
}


void drawborder ()
{
	PortSet (light_plane, 239, 127);
	ClrScr();
	DrawClipRect (MakeWinRect (0, 0, 8*(Xsize+1)-1, 8*(Ysize+1)-1), full_screen, A_NORMAL | B_NORMAL);
	
	PortSet (dark_plane, 239, 127);
	ClrScr();
	DrawClipRect (MakeWinRect (0, 0, 8*(Xsize+1)-1, 8*(Ysize+1)-1), full_screen, A_NORMAL | B_DOUBLE);
	
	PortRestore();
}

void drawblock (int x, int y, int type)
{
	int xpos, ypos;
	xpos= 4+8*x;
	ypos= 4+8*y;
	
	MySprite8 (xpos, ypos, BL_sprites+16*type);
	
	return;
}

void drawlevel ()
{
	void drawborder ();
	int readblock (int x, int y);
	int I,J;
	
	drawborder ();
	for (I=0; I<Xsize; I++)
	{
		for (J=0; J<Ysize; J++) { drawblock (I, J, readblock (I, J)); }
	}
		
}


int readblock (int x, int y)
{
	unsigned int temp= *(level+ y*((Xsize+1)/2)+ x/2);
	
	if (x<0 || x>(Xsize-1) || y<0 || y>(Ysize-1)) { return -1; }
	if (!(x%2)) { temp>>=4; }
	return temp%16;
}


int bl_quit()
{
	void loadlevel();
	void drawlevel ();
	void drawchar (int type);
		
	if (quit())
	{
		int old_Xscroll= Xscroll, old_Yscroll= Yscroll;
		int retval;
		Xscroll= 0;
		Yscroll= 0;
		while (quit()) {}
		retval= intro_screen();
		Xscroll= old_Xscroll;
		Yscroll= old_Yscroll;
		drawlevel ();
		drawchar (0);
		return retval;
	}
	else if ( ((!is_92) && (rowread_func(0xfd)&0x40)) || (is_92 && (rowread_func(0xfe)&0x08)) ) { loadlevel(); }
	else if ((!to_win) || (rowread_func(0xfd)&0x02))
		{
			PortSet (dark_plane, 239, 99);
			FontSetSys (F_8x10);
			print_string (100, 60, "You Win");
//			DlgMessage (NULL, "You win!!", BT_OK, BT_NONE);
			PortRestore ();
			waitkey();
			Curr_level+= (((Xsize+1)/2)*Ysize+HEADER_SIZE);
			loadlevel();
			if (Xsize==0) {return 1;}
		}
	return 0;
}


void scroll()
{
	int key= rowread_func(0xfffe);

	if (Xsize<20 && Ysize<12) { return; }

	if ((key&0x0f) && !keydown && (key&0x10))
	{
		if (key&0x8 && Xscroll<(8*Xsize-152)) { Xscroll++; }
		if (key&0x2 && Xscroll) { Xscroll--; }
		if (key&0x4 && Yscroll<(8*Ysize-92)) { Yscroll++; }
		if (key&0x1 && Yscroll) { Yscroll--; }
		keydown= 1;
	}
}


void getmove()
{
	void move (int mymove);
	//enum type_move {MV_UP, MV_RIGHT, MV_DOWN, MV_LEFT};	
	
	int key= getkey();
	if (key && (key<5) && (!movekeydown) && !( (!is_92) && (rowread_func(0xfffe)&0x10))  )
	{
		move (key-1);
		movekeydown= 1;
	}
	else if ((!key) && movekeydown)
	{
		movekeydown= 0;
	}
	
}


void move (int mymove)
{
	//enum type_move {MV_UP, MV_RIGHT, MV_DOWN, MV_LEFT};
	//enum blocks { BL_VOID, BL_FIXE, BL_LEGER, BL_HEAVY, BL_HOLE, BL_REACT, BL_UP, BL_RIGHT, BL_DOWN, BL_LEFT, BL_HORIZ, BL_VERT, BL_EXPLODE, BL_JUMP, BL_STEEL, BL_WIN };
	
	typedef union
	{
		struct { unsigned char move_to: 1, move_up: 1, move_right: 1, move_down: 1, move_left: 1, movem: 1, explodable: 1, fallable: 1; } bits;
		unsigned char all;
	} bl_def;
	
	
	int nextX= Xpos, newX= Xpos, nextY= Ypos, newY= Ypos;
	int block, nextblock;
	
//	extern bl_def* def;
			
	bl_def def[16]= {	{1, 0, 0, 0, 0, 0, 0, 0},	//Void
				{0, 0, 0, 0, 0, 0, 1, 0},	//Fixe
				{0, 1, 1, 1, 1, 1, 1, 1},	//Leger
				{0, 1, 1, 1, 1, 0, 1, 1},	//Heavy
				{0, 0, 0, 0, 0, 0, 0, 0},	//Hole
				{0, 1, 1, 1, 1, 0, 1, 0},	//React
				{0, 1, 0, 0, 0, 0, 1, 1},	//Up
				{0, 0, 1, 0, 0, 0, 1, 1},	//Right
				{0, 0, 0, 1, 0, 0, 1, 1},	//Down
				{0, 0, 0, 0, 1, 0, 1, 1},	//Left
				{0, 0, 1, 0, 1, 0, 1, 1},	//Horiz
				{0, 1, 0, 1, 0, 0, 1, 1},	//Vert
				{0, 1, 1, 1, 1, 0, 1, 0},	//Explode
				{1, 0, 0, 0, 0, 0, 1, 0},	//Jump
				{0, 0, 0, 0, 0, 0, 0, 0},	//Steel
				{1, 0, 0, 0, 0, 0, 0, 0}	//Win
					};
	
/*	bl_def def[16]= {	0x80,	//Void
				0x02,	//Fixe
				0x7f,	//Leger
				0x7b,	//Heavy
				0x00,	//Hole
				0x7e,	//React
				0x43,	//Up
				0x23,	//Right
				0x13,	//Down
				0x0b,	//Left
				0x2b,	//Horiz
				0x53,	//Vert
				0x7a,	//Explode
				0x82,	//Jump
				0x02,	//Teleport
				0x80	//Win
					};*/
	
	void writelevel (int x, int y, int type);
	int readblock (int x, int y);
	void movechar (int newX, int newY);
	void do_move (int mymove,int* x, int* y)
	{
		switch (mymove)
		{
			case MV_UP:
				(*y)--;
				break;
			case MV_RIGHT:
				(*x)++;
				break;
			case MV_DOWN:
				(*y)++;
				break;
			case MV_LEFT:
				(*x)--;
		}
	}
	
	
	
	do_move (mymove, &newX, &newY);
	nextX= newX; nextY= newY;
	do_move (mymove, &nextX, &nextY);
	
	block= readblock (newX, newY);
	if (block<0) { return; }
	if (def[block].bits.move_to)
	{
		if (block==BL_JUMP)
		{
			do_move (mymove, &newX, &newY);
			do_move (mymove, &newX, &newY);
			block= readblock(newX, newY);
			if ((block<0) || (!def[block].bits.move_to)) { return; }
		}
		if (block==BL_WIN)
		{
			writelevel (newX, newY, BL_VOID);
			to_win--;
		}
		movechar (newX, newY);
		return;
	}
	if ( (def[block].all>>(6-mymove))%2 ) //if move is allowed in the selected direction
	{
		if (block==BL_REACT)
		{
			nextX=Xpos; nextY=Ypos;
			do_move ((mymove+2)%4, &nextX, &nextY);
			nextblock= readblock(nextX, nextY);
			if (nextblock<0) { return; }
			if (nextblock==BL_VOID)
			{
				writelevel (newX, newY, BL_VOID);
				writelevel (Xpos, Ypos, BL_REACT);
				movechar (nextX, nextY);
			}
			return;
		}
		nextblock= readblock(nextX, nextY);
		if (nextblock<0) { return; }
		if (def[block].bits.movem && def[nextblock].bits.movem)
		{
			while (def[nextblock].bits.movem && nextblock>=0 )
			{ do_move (mymove, &nextX, &nextY); nextblock= readblock(nextX, nextY); }			
		}
		if (nextblock==BL_VOID)
		{
			writelevel (newX, newY, BL_VOID);
			writelevel (nextX, nextY, block);
			movechar (newX, newY);
			return;
		}
		if (	(nextblock==BL_HOLE && (def[block].bits.fallable))
		   ||	((block==BL_EXPLODE) && (def[nextblock].bits.explodable))  )
		{
			writelevel (newX, newY, BL_VOID);
			writelevel (nextX, nextY, BL_VOID);
			movechar (newX, newY);
			return;
		}
	}
}


void writelevel(int x, int y, int type)
{
	void drawblock (int x, int y, int type);
	
	unsigned int temp;
	temp= 0xf;
	if (!(x%2)) { temp<<=4; }
	temp^= 0xff;
	level[y*((Xsize+1)/2)+ (x/2)] &=temp;	//efface
	temp= type;
	if (!(x%2)) { temp<<=4; }
	level[y*((Xsize+1)/2)+ (x/2)] |=temp;	//ecrit
	
	drawblock (x, y, type);
}

	
void drawchar (int type)
{
	int xpos, ypos;
	xpos= 4+8*Xpos;
	ypos= 4+8*Ypos;
	
	MySprite8 (xpos, ypos, CHAR_sprites+16*type);
	
	return;
}


void movechar (int newX, int newY)
{
	drawblock (Xpos, Ypos, readblock (Xpos, Ypos));
	Xpos= newX;
	Ypos= newY;
	if (readblock (newX,newY)==BL_JUMP) { drawchar (1); }
	else { drawchar (0); }
}